/* Revision History
 * 2015-02-11 - MisterCardholder
 *     The script can load most models with UV maps, vertex normals, basic 
 *     textures, armatures, and bone weights.
 * 2015-02-12 - MisterCardholder
 *     Name textures to make importing into other software easier
 *     Added support for accessing the second set of meshes used in some files
 */

/* Notes:
 *
 * When creating materials, the script currently ignores everything except for
 * the albedo and normal maps.
 *
 * Many characters come in several pieces, each with their own wmb file.  In
 * most cases the textures are only stored with the first wmb, so subsequent
 * pieces will often be texture-less.
 *
 * When exporting stages, Noesis seems to export one model at a time, but 
 * includes <numberOfModelsInStage> copies of each texture, whether they're 
 * used in that particular model or not.  I'm not sure if that's because I'm
 * doing something wrong or if that's just how Noesis works when dealing with
 * multiple models.
 *
 */

#include "stdafx.h"
#include <stddef.h>
#include "wmb4.h"
#include "wta.h"
#include "scr.h"

const char *g_pPluginName = "MetalGearRisingRevengeancePC";
const char *g_pPluginDesc = "Metal Gear Rising: Revengeance (PC) model handler, by MisterCardholder";

/// Determines if we can handle the given file
bool mgrrTypeCheckModel(BYTE *fileBuffer, int bufferLen, noeRAPI_t *rapi) {
	if (bufferLen < sizeof(wmb4_header_t)) {
		// Too small to be a WMB file
		return false;
	}
	wmb4_header_t *fileHeader = ((wmb4_header_t *)fileBuffer);
	if (memcmp(&(fileHeader->fourCC), "WMB4", 4)) {
		// WMB4 header missing
		return false;
	}
	// All signs point to a valid file
	return true;
}

/// Determines if we can handle the given file
bool mgrrTypeCheckStage(BYTE *fileBuffer, int bufferLen, noeRAPI_t *rapi) {
	if (bufferLen < sizeof(scr_header_t)) {
		// Too small to be a SCR file
		return false;
	}
	scr_header_t *fileHeader = ((scr_header_t *)fileBuffer);
	if (memcmp(&(fileHeader->fourCC), "SCR\0", 4)) {
		// SCR header missing
		return false;
	}
	uint32_t modelHeaderCount = fileHeader->modelHeaderCount;
	if (modelHeaderCount == 0) {
		// There don't seem to be any models in here
		return false;
	}
	if (bufferLen < fileHeader->modelHeaderOffsetArrayOffset) {
		// File too small
		return false;
	}
	scr_modelHeaderOffset_t *modelHeaderOffsetBuffer = (scr_modelHeaderOffset_t *)(fileBuffer + fileHeader->modelHeaderOffsetArrayOffset);
	if (bufferLen < modelHeaderOffsetBuffer[0]) {
		// File too small
		return false;
	}
	scr_modelHeader_t *modelHeader = (scr_modelHeader_t *)(fileBuffer + modelHeaderOffsetBuffer[0]);
	if (bufferLen < modelHeader->wmbOffset) {
		// File too small
		return false;
	}
	BYTE *wmbBuffer = fileBuffer + modelHeader->wmbOffset;
	if (memcmp(wmbBuffer, "WMB4", 4)) {
		// WMB4 header missing
		return false;
	}
	// All signs point to a valid file
	return true;
}

void unpackNormals(uint32_t packed, float *x, float *y, float *z) {
	int xWidth = 11;
	int yWidth = 11;
	int zWidth = 10;
	int xPos = 0;
	int yPos = xPos + xWidth;
	int zPos = yPos + yWidth;
	size_t fullWidth = sizeof(packed) * 8;

	int xInt = ((int)(packed << (fullWidth - xPos - xWidth))) >> (fullWidth - xWidth);
	int yInt = ((int)(packed << (fullWidth - yPos - yWidth))) >> (fullWidth - yWidth);
	int zInt = ((int)(packed << (fullWidth - zPos - zWidth))) >> (fullWidth - zWidth);

	*x = (xInt) / (float)((1 << (xWidth - 1)) - 1);
	*y = (yInt) / (float)((1 << (yWidth - 1)) - 1);
	*z = (zInt) / (float)((1 << (zWidth - 1)) - 1);
}

// You must free the pointer returned by Noesis_ReadFile with Noesis_UnpooledFree! (unless it's NULL)
BYTE *mgrrLoadRelatedFile(char *extension, noeRAPI_t *rapi, int &filesize) {
	char *inFileName = rapi->Noesis_GetInputName();
	BYTE *relatedFile = NULL;
	filesize = 0;
	if (inFileName && inFileName[0]) {
		char relatedFileName[MAX_NOESIS_PATH];
		rapi->Noesis_GetExtensionlessName(relatedFileName, inFileName);
		strcat_s(relatedFileName, MAX_NOESIS_PATH, extension);

		int relatedFileLength = 0;
		relatedFile = (BYTE *)rapi->Noesis_ReadFile(relatedFileName, &relatedFileLength);
		filesize = relatedFileLength;

		if (!relatedFile && rapi->Noesis_CheckFileExt(inFileName, ".scr")) {
			// The naming scheme for wta/wtp files associated with scr files
			// can be slightly different: xxx.scr -> xxxscr.wta NOT xxx.wta
			rapi->Noesis_GetExtensionlessName(relatedFileName, inFileName);
			strcat_s(relatedFileName, MAX_NOESIS_PATH, "scr");
			strcat_s(relatedFileName, MAX_NOESIS_PATH, extension);

			int relatedFileLength = 0;
			relatedFile = (BYTE *)rapi->Noesis_ReadFile(relatedFileName, &relatedFileLength);
			filesize = relatedFileLength;
		}
	}
	return relatedFile;
}

bool mgrrValidateWTAFile(BYTE *fileBuffer, int bufferLen) {
	if (bufferLen < sizeof(wta_header_t)) {
		// Too small to be a WTA file
		return false;
	}
	wta_header_t *fileHeader = ((wta_header_t *)fileBuffer);
	if (memcmp(&(fileHeader->fourCC), "WTB\0", 4)) {
		// WTA header missing
		return false;
	}
	return true;
}

void mgrrLoadTextures(CArrayList<noesisTex_t *> &textures, noeRAPI_t *rapi) {
	int wtaFileLength = 0;
	BYTE *wtaFile = NULL;
	int wtpFileLength = 0;
	BYTE *wtpFile = NULL;

	wtaFile = mgrrLoadRelatedFile(".wta", rapi, wtaFileLength);
	if (wtaFile) {
		wtpFile = mgrrLoadRelatedFile(".wtp", rapi, wtpFileLength);
	} else {
		wtaFile = mgrrLoadRelatedFile(".wtb", rapi, wtaFileLength);
		wtpFile = wtaFile;
		wtpFileLength = wtaFileLength;
	}

	if (mgrrValidateWTAFile(wtaFile, wtaFileLength)) {
		wta_header_t *wtaHeader = (wta_header_t *)wtaFile;
		uint32_t textureCount = wtaHeader->textureCount;
		wta_textureOffset_t *textureOffsetBuffer = (wta_textureOffset_t *)(wtaFile + wtaHeader->textureOffsetArrayOffset);
		wta_textureSize_t *textureSizeBuffer = (wta_textureSize_t *)(wtaFile + wtaHeader->textureSizeArrayOffset);
		wta_textureIdentifier_t *textureIdentifierBuffer = (wta_textureIdentifier_t *)(wtaFile + wtaHeader->textureIdentifierArrayOffset);

		for (int i = 0; i < textureCount; i++) {
			noesisTex_t *texture = NULL;
			if ((textureOffsetBuffer[i] + textureSizeBuffer[i]) <= wtpFileLength) {
				texture = rapi->Noesis_LoadTexByHandler((wtpFile + textureOffsetBuffer[i]), textureSizeBuffer[i], ".dds");
			}
			if (!texture) {
				texture = rapi->Noesis_AllocPlaceholderTex(NULL, 32, 32, false);
				rapi->LogOutput("WARNING: Texture %d not found in the WTP file.\n", i);
			}
			texture->globalIdx = textureIdentifierBuffer[i];

			char texName[128];
			sprintf_s(texName, 128, "%08X", textureIdentifierBuffer[i]);
			texture->filename = rapi->Noesis_PooledString(texName);

			textures.Append(texture);
		}
	}

	if (wtaFile) {
		if (wtpFile == wtaFile) {
			wtpFile = NULL;
		}
		rapi->Noesis_UnpooledFree(wtaFile);
	}
	if (wtpFile) {
		rapi->Noesis_UnpooledFree(wtpFile);
	}
}

// Searches the list of textures for one with the given identifier
// Returns -1 if one cannot be found
int mgrrIndexOfTextureWithIdentifier(CArrayList<noesisTex_t *> &textures, uint32_t textureIdentifier) {
	int textureIndex = -1;
	for (int k = 0; k < textures.Num(); k++) {
		noesisTex_t *tex = textures[k];
		if (tex && (tex->globalIdx == textureIdentifier)) {
			textureIndex = k;
			break;
		}
	}
	return textureIndex;
}

void mgrrCreateMaterials(wmb4_header_t &fileHeader, BYTE *fileBuffer, CArrayList<noesisMaterial_t *> &materials, CArrayList<noesisTex_t *> &textures, noeRAPI_t *rapi) {
	wmb4_textureIdentifier_t *textureIdentifierBuffer = (wmb4_textureIdentifier_t *)(fileBuffer + fileHeader.textureIdentifierArrayOffset);
	uint32_t textureIdentifierCount = fileHeader.textureIdentifierCount;

	wmb4_material_t *materialBuffer = (wmb4_material_t *)(fileBuffer + fileHeader.materialArrayOffset);
	uint32_t materialCount = fileHeader.materialCount;

	// For each material...
	for (int i = 0; i < materialCount; i++) {
		noesisMaterial_t *noesisMaterial = rapi->Noesis_GetMaterialList(1, false);
		noesisMaterial->texIdx = -1;

		wmb4_textureMap_t *textureMapBuffer = (wmb4_textureMap_t *)(fileBuffer + materialBuffer[i].textureMapArrayOffset);

		uint32_t textureIdentifierIndex;
		
		// Albedo
		textureIdentifierIndex = textureMapBuffer->albedoMapTextureIndex;
		if (textureIdentifierIndex < textureIdentifierCount) {
			uint32_t textureIdentifier = textureIdentifierBuffer[textureIdentifierIndex].textureIdentifier;
			int32_t index = mgrrIndexOfTextureWithIdentifier(textures, textureIdentifier);
			if (index >= 0) {
				noesisMaterial->texIdx = index;
			} else {
				// We know the texture's identifier, but it wasn't found in memory!
				noesisTex_t *noesisTexture = rapi->Noesis_AllocPlaceholderTex(NULL, 32, 32, false);
				char texName[128];
				sprintf_s(texName, 128, "%08X", textureIdentifier);
				noesisTexture->filename = rapi->Noesis_PooledString(texName);

				noesisMaterial->texIdx = textures.Num();
				textures.Append(noesisTexture);
				rapi->LogOutput("WARNING: The an albedo texture for material %d was not found in memory.\n", i);
			}
		} else {
			// The material specified an invalid texture; this should be impossible
			noesisTex_t *noesisTexture = rapi->Noesis_AllocPlaceholderTex(NULL, 32, 32, false);
			noesisMaterial->texIdx = textures.Num();
			textures.Append(noesisTexture);
			rapi->LogOutput("WARNING: Failed to load an albedo texture for material %d.\n", i);
		}

		// Normal
		textureIdentifierIndex = textureMapBuffer->normalMapTextureIndex;
		if (textureIdentifierIndex < textureIdentifierCount) {
			uint32_t textureIdentifier = textureIdentifierBuffer[textureIdentifierIndex].textureIdentifier;
			int32_t index = mgrrIndexOfTextureWithIdentifier(textures, textureIdentifier);
			if (index >= 0) {
				noesisMaterial->normalTexIdx = index;
			}
			else {
				// We know the texture's identifier, but it wasn't found in memory!
				noesisTex_t *noesisTexture = rapi->Noesis_AllocPlaceholderTex(NULL, 32, 32, false);
				char texName[128];
				sprintf_s(texName, 128, "%08X", textureIdentifier);
				noesisTexture->filename = rapi->Noesis_PooledString(texName);

				noesisMaterial->normalTexIdx = textures.Num();
				textures.Append(noesisTexture);
				rapi->LogOutput("WARNING: Failed to load an albedo texture for material %d.\n", i);
			}
		}

		char matName[128];
		sprintf_s(matName, 128, "material%03d", i);
		noesisMaterial->name = rapi->Noesis_PooledString(matName);

		materials.Append(noesisMaterial);
	}
}

modelBone_t *mgrrCreateBones(wmb4_header_t &fileHeader, BYTE *fileBuffer, noeRAPI_t *rapi, int &numBones) {
	numBones = 0;
	if (fileHeader.boneCount == 0) {
		return NULL;
	}
	wmb4_bone_t *boneBuffer = (wmb4_bone_t *)(fileBuffer + fileHeader.boneArrayOffset);
	numBones = fileHeader.boneCount;

	modelBone_t *bones = rapi->Noesis_AllocBones(numBones);

	for (int i = 0; i < numBones; i++) {
		modelBone_t *bone = bones + i;

		sprintf_s(bone->name, 30, "bone%03d", boneBuffer[i].boneNumber);

		short parent = boneBuffer[i].parentIndex;
		assert(parent < numBones);
		bone->eData.parent = (parent >= 0) ? &(bones[parent]) : NULL;

		bone->mat = g_identityMatrix;
		float pos[3];
		memcpy(pos, &(boneBuffer[i].positionAbsoluteX), sizeof(pos));
		g_mfn->Math_VecCopy(pos, bone->mat.o);
	}

	return bones;
}

noesisModel_t *mgrrCreateModel(BYTE *fileBuffer, CArrayList<noesisTex_t *> &textures, noeRAPI_t *rapi) {

	wmb4_header_t *fileHeader = (wmb4_header_t *)fileBuffer;

	wmb4_vertexHeader_t *vertexHeaderBuffer = (wmb4_vertexHeader_t *)(fileBuffer + fileHeader->vertexHeaderArrayOffset);

	uint32_t vertexStride;
	int32_t vertexPositionOffset;
	int32_t vertexUVOffset;
	int32_t vertexNormalsOffset;
	int32_t boneIndexOffset;
	int32_t boneWeightOffset;

	// This is almost certainly wrong, but it seems to work in most cases. 
	if ((fileHeader->vertexFlag1 == 0x37) || (fileHeader->vertexFlag2 == 0x03)) {
		vertexStride 			= sizeof(wmb4_vertex32_t);
		vertexPositionOffset 	= offsetof(wmb4_vertex32_s, positionX);
		vertexUVOffset 			= offsetof(wmb4_vertex32_s, textureU);
		vertexNormalsOffset 	= offsetof(wmb4_vertex32_s, normals);
		boneIndexOffset 		= offsetof(wmb4_vertex32_t, boneIndex0);
		boneWeightOffset 		= offsetof(wmb4_vertex32_t, boneWeight0);
	} else if (fileHeader->vertexFlag2 == fileHeader->vertexFlag3 == 0x01) {
		vertexStride 			= sizeof(wmb4_vertex28_t);
		vertexPositionOffset 	= offsetof(wmb4_vertex28_s, positionX);
		vertexUVOffset 			= offsetof(wmb4_vertex28_s, textureU);
		vertexNormalsOffset 	= offsetof(wmb4_vertex28_s, normals);
		boneIndexOffset 		= -1;
		boneWeightOffset 		= -1;
	} else if ((fileHeader->vertexFlag2 == 0x01) && (fileHeader->vertexFlag3 == 0x00)) {
		vertexStride			= sizeof(wmb4_vertex24_t);
		vertexPositionOffset	= offsetof(wmb4_vertex24_s, positionX);
		vertexUVOffset			= offsetof(wmb4_vertex24_s, textureU);
		vertexNormalsOffset		= offsetof(wmb4_vertex24_s, normals);
		boneIndexOffset			= -1;
		boneWeightOffset		= -1;
	} else {
		// Unknown vertex format
		return NULL;
	}
	
	wmb4_mesh_t *meshBuffer = (wmb4_mesh_t *)(fileBuffer + fileHeader->meshArrayOffset);
	uint32_t meshCount = fileHeader->meshCount;

	wmb4_meshMetadataHeader_t *meshMetadataHeader = (wmb4_meshMetadataHeader_t *)(fileBuffer + fileHeader->meshMetadataHeaderOffset);
	wmb4_meshMetadata_t *meshMetadataBuffer = (wmb4_meshMetadata_t *)(fileBuffer + meshMetadataHeader->meshMetadataOffset);
	uint32_t meshMetadataCount = meshMetadataHeader->meshMetadataCount;
	wmb4_meshMetadata_t *altMeshMetadataBuffer = (wmb4_meshMetadata_t *)(fileBuffer + meshMetadataHeader->altMeshMetadataOffset);
	uint32_t altMeshMetadataCount = meshMetadataHeader->altMeshMetadataCount;

	wmb4_meshNameHeader_t *meshNameHeaderBuffer = (wmb4_meshNameHeader_t *)(fileBuffer + fileHeader->meshNameHeaderArrayOffset);
	uint32_t meshNameHeaderCount = fileHeader->meshNameHeaderCount;

	wmb4_boneMap_t *boneMapBuffer = (wmb4_boneMap_t *)(fileBuffer + fileHeader->boneMapOffset);
	uint32_t boneMapCount = fileHeader->boneMapCount;

	CArrayList<noesisMaterial_t *> materials;
	mgrrCreateMaterials(*fileHeader, fileBuffer, materials, textures, rapi);

	int numBones;
	modelBone_t *bones = mgrrCreateBones(*fileHeader, fileBuffer, rapi, numBones);


	void *pgctx = rapi->rpgCreateContext();

	// For each mesh...
	for (int i = 0; i < meshCount; i++) {
		// Find the mesh's metadata
		wmb4_meshMetadata_t *metadata = NULL;
		bool isAlternateMesh = false;
		for (int j = 0; j < meshMetadataCount; j++) {
			if ((meshMetadataBuffer + j)->meshIndex == i) {
				metadata = (meshMetadataBuffer + j);
				isAlternateMesh = false;
				break;
			}
		}
		if (!metadata) {
			for (int j = 0; j < altMeshMetadataCount; j++) {
				if ((altMeshMetadataBuffer + j)->meshIndex == i) {
					metadata = (altMeshMetadataBuffer + j);
					isAlternateMesh = true;
					break;
				}
			}
		}
		if (metadata == NULL) {
			// Ignore meshes without metadata
			rapi->LogOutput("WARNING: Ignoring mesh %d.  No metadata found.\n", i);
			continue;
		}

		// Set the mesh's name
		uint32_t meshNameIndex = metadata->meshNameIndex;
		if (meshNameIndex < meshNameHeaderCount) {
			char meshName[64];
			sprintf(meshName, "mesh%02d-%s", i, (char *)(fileBuffer + meshNameHeaderBuffer[meshNameIndex].nameOffset));
			rapi->rpgSetName(meshName);
		} else {
			char meshName[64];
			sprintf(meshName, "mesh%02d", i);
			rapi->rpgSetName(meshName);
		}

		// Find the vertex header

		void *vertexBuffer = (fileBuffer + vertexHeaderBuffer[meshBuffer[i].vertexHeaderIndex].vertexArrayOffset);
		uint32_t vertexCount = vertexHeaderBuffer[meshBuffer[i].vertexHeaderIndex].vertexCount;

		wmb4_face_t *faceBuffer = (wmb4_face_t *)(fileBuffer + vertexHeaderBuffer[meshBuffer[i].vertexHeaderIndex].faceArrayOffset);
		uint32_t faceCount = vertexHeaderBuffer[meshBuffer[i].vertexHeaderIndex].faceCount;

		// Get the bone reference map
		int *boneReferenceMap = NULL;
		uint32_t boneMapIndex = metadata->boneMapIndex;
		if (boneMapIndex < boneMapCount) {
			wmb4_boneMap_t *boneMap = &(boneMapBuffer[boneMapIndex]);
			wmb4_boneMapEntry_t *boneMapEntries = (wmb4_boneMapEntry_t *)(fileBuffer + boneMap->boneMapOffset);
			boneReferenceMap = (int *)rapi->Noesis_UnpooledAlloc(sizeof(int)*boneMap->boneMapCount);
			for (int j = 0; j < boneMap->boneMapCount; j++) {
				boneReferenceMap[j] = boneMapEntries[j];
			}
			rapi->rpgSetBoneMap(boneReferenceMap);
		}

		// Set the material
		uint32_t materialIndex = metadata->materialIndex;
		rapi->rpgSetMaterialIndex(materialIndex);

		rapi->rpgBegin(RPGEO_TRIANGLE);

		// For each (triangular) face...
		for (int j = 0; j < meshBuffer[i].faceCount; j += 3) {
			int faceBufferIndex = meshBuffer[i].faceStart + j;
			int vertexStartIndex = meshBuffer[i].vertexStart;	// The first vertex in this mesh

			// For each vertex...
			for (int k = 0; k < 3; k++) {
				int vertexBufferIndex = vertexStartIndex + faceBuffer[faceBufferIndex + k];

				if ((boneIndexOffset >= 0) && (boneWeightOffset >= 0)) {
					rapi->rpgVertBoneIndexUB((BYTE *)((uint8_t *)vertexBuffer + boneIndexOffset + (vertexBufferIndex * vertexStride)), 4);
					rapi->rpgVertBoneWeightUB((BYTE *)((uint8_t *)vertexBuffer + boneWeightOffset + (vertexBufferIndex * vertexStride)), 4);
				}

				uint32_t packedNormals = *((uint32_t *)((uint8_t *)vertexBuffer + vertexNormalsOffset + (vertexBufferIndex * vertexStride)));
				float normals[3];
				unpackNormals(packedNormals, normals, normals + 1, normals + 2);
				rapi->rpgVertNormal3f(normals);

				rapi->rpgVertUV2hf((WORD *)((uint8_t *)vertexBuffer + vertexUVOffset + (vertexBufferIndex * vertexStride)), 0);
				rapi->rpgVertex3f((float *)((uint8_t *)vertexBuffer + vertexPositionOffset + (vertexBufferIndex * vertexStride)));
			}
		}

		rapi->rpgEnd();

		if (boneReferenceMap) {
			rapi->rpgSetBoneMap(NULL);
			rapi->Noesis_UnpooledFree(boneReferenceMap);
		}

		rapi->rpgSetMaterialIndex(-1);
	}

	noesisMatData_t *md = rapi->Noesis_GetMatDataFromLists(materials, textures);
	rapi->rpgSetExData_Materials(md);

	rapi->rpgSetExData_Bones(bones, numBones);

	noesisModel_t *mdl = rapi->rpgConstructModel();

	rapi->rpgDestroyContext(pgctx);

	return mdl;
}

noesisModel_t *mgrrLoadModel(BYTE *fileBuffer, int bufferLen, int &numMdl, noeRAPI_t *rapi) {
	CArrayList<noesisTex_t *> textures;
	mgrrLoadTextures(textures, rapi);

	noesisModel_t *mdl = mgrrCreateModel(fileBuffer, textures, rapi);
	if (mdl) {
		numMdl = 1;
	}

	return mdl;
}

noesisModel_t *mgrrLoadStage(BYTE *fileBuffer, int bufferLen, int &numMdl, noeRAPI_t *rapi) {
	scr_header_t *fileHeader = (scr_header_t *)fileBuffer;

	scr_modelHeaderOffset_t *modelHeaderOffsetBuffer = (scr_modelHeaderOffset_t *)(fileBuffer + fileHeader->modelHeaderOffsetArrayOffset);
	uint32_t modelHeaderCount = fileHeader->modelHeaderCount;
	
	CArrayList<noesisTex_t *> textures;
	mgrrLoadTextures(textures, rapi);

	CArrayList<noesisModel_t *> models;
	for (int i = 0; i < modelHeaderCount; i++) {
		scr_modelHeader_t *modelHeader = (scr_modelHeader_t *)(fileBuffer + modelHeaderOffsetBuffer[i]);
		BYTE *wmbBuffer = fileBuffer + modelHeader->wmbOffset;

		noesisModel_t *mdl = mgrrCreateModel(wmbBuffer, textures, rapi);
		if (mdl) {
			models.Append(mdl);
		}
	}

	if (models.Num() <= 0) {
		return NULL;
	}
	noesisModel_t *mdlList = rapi->Noesis_ModelsFromList(models, numMdl);
	models.Clear();

	return mdlList;
}

//called by Noesis to init the plugin
bool NPAPI_InitLocal(void) {
	if (g_nfn->NPAPI_GetAPIVersion() < NOESIS_PLUGINAPI_VERSION) {
		return false;
	}

	int fh = g_nfn->NPAPI_Register("Metal Gear Rising: Revengeance (PC) Model", ".wmb");
	if (fh < 0)	{
		return false;
	}

	//set the data handlers for this format
	g_nfn->NPAPI_SetTypeHandler_TypeCheck(fh, mgrrTypeCheckModel);
	g_nfn->NPAPI_SetTypeHandler_LoadModel(fh, mgrrLoadModel);


	fh = g_nfn->NPAPI_Register("Metal Gear Rising: Revengeance (PC) Stage", ".scr");
	if (fh < 0)	{
		return false;
	}

	//set the data handlers for this format
	g_nfn->NPAPI_SetTypeHandler_TypeCheck(fh, mgrrTypeCheckStage);
	g_nfn->NPAPI_SetTypeHandler_LoadModel(fh, mgrrLoadStage);

	return true;
}

//called by Noesis before the plugin is freed
void NPAPI_ShutdownLocal(void) {
	// nothing to do here
}

BOOL APIENTRY DllMain(HMODULE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved) {
	return TRUE;
}

