/* Revision History
 * 2015-02-11 - MisterCardholder
 *   Initial Release
 * 2015-02-12 - MisterCardholder
 *   Updates to wmb4_meshMetadataHeader_t
 */

/*
 * Metal Gear Rising: Revengeance (PC) WMB File Structure
 *
 * +-------------------------------------+
 * | File Header                         |
 * | wmb4_header_t                       |
 * +-------------------------------------+
 * | Array of Vertex Headers             |
 * | wmb4_vertexHeader_t                 |
 * |   ----------------------------------+
 * |   Array of Vertexes                 |
 * |   wmb4_vertexXX_t                   |
 * |   ----------------------------------+
 * |   Array of Unknown Data             |
 * |                                     |
 * |   ----------------------------------+
 * |   Array of Faces                    |
 * |   wmb4_face_t                       |
 * +-------------------------------------+
 * | Array of Meshes                     |
 * | wmb4_mesh_t                         |
 * +-------------------------------------+
 * | Mesh Metadata Header                |
 * | wmb4_meshMetadataHeader_t           |
 * |   ----------------------------------+
 * |   Array of Visible Mesh Metadata    |
 * |   wmb4_meshMetadata_t               |
 * |   ----------------------------------+
 * |   Array of Hidden Mesh Metadata     |
 * |   wmb4_meshMetadata_t               |
 * +-------------------------------------+
 * | Array of Bones                      |
 * | wmb4_bone_t                         |
 * +-------------------------------------+
 * | Array of Unknown Data               |
 * | Appears to be an array of uint8_t's |
 * | I've seen a suggestion that they    |
 * | might be normalized floats (so      |
 * | divide by 0xFF to get a value [0,1])|
 * | but their purpose is unclear.       |
 * +-------------------------------------+
 * | Array of Bone Map Offsets           |
 * | wmb4_boneMap_t                      |
 * |   ----------------------------------+
 * |   Array of Bone Map Indexes         |
 * |   uint16_t                          |
 * +-------------------------------------+ 
 * | Array of Materials                  |
 * | wmb4_material_t                     |
 * |   ----------------------------------+
 * |   Materials Data                    |
 * |   wmb4_textureMap_t                 |
 * +-------------------------------------+
 * | Array of Texture Identifiers        | 
 * | wmb4_textureIdentifier_t            |
 * +-------------------------------------+
 * | Array of Mesh Names Headers         |
 * | wmb4_meshNameHeader_t               |
 * |   ----------------------------------+
 * |   Mesh Name Data                    |
 * |                                     |
 * +-------------------------------------+
 *
 * Notes:
 * 
 * All offsets are from the start of the file.
 *
 * Many sections of the file are zero-padded to be 16-byte-aligned, so
 * there may be instances where structures either mistakenly include those 
 * zeros or omit fields because they happened to be zero when documenting the
 * file structure.
 *
 */

#include <stdint.h>

typedef struct wmb4_header_s {
	char		fourCC[4]; 	// "WMB4"
	uint32_t	unknown04;	// 0x00000000
	uint8_t		vertexFlag1;	// + 
	uint8_t		vertexFlag2;	// | These indicate the vertex format used, but
	uint8_t		vertexFlag3;	// | the details are currently unknown.
	uint8_t		vertexFlag4;	// + 
	uint16_t	unknown0C;		// 0x0000
	uint16_t	unknown0E;		// 0xFFFF (often but not always)
	float		unknown10;
	float		unknown14;
	float		unknown18;
	float		unknown1C;
	float		unknown20;
	float		unknown24;
	uint32_t	vertexHeaderArrayOffset;		// 0x28 see: wmb4_vertexHeader_t
	uint32_t	vertexHeaderArrayCount;			// This may actually not be a count...
	uint32_t	meshArrayOffset;				// 0x30 see: wmb4_mesh_t
	uint32_t	meshCount;
	uint32_t	meshMetadataHeaderOffset;		// 0x38 see: wmb4_meshMetadataHeader_t
	uint32_t	boneArrayOffset;				// 0x3C see: wmb4_bone_t
	uint32_t	boneCount;
	uint32_t	unknown44;						// Offset
	uint32_t	unknown48;						// Count
	uint32_t	boneMapOffset;					// 0x4C see wmb4_boneMap_t
	uint32_t	boneMapCount;
	uint32_t	materialArrayOffset;			// 0x54 see wmb4_material_t
	uint32_t	materialCount;
	uint32_t	textureIdentifierArrayOffset;	// 0x5C see wmb4_textureIdentifier_t
	uint32_t	textureIdentifierCount;
	uint32_t	meshNameHeaderArrayOffset;		// 0x64 see: wmb4_meshNameHeader_t
	uint32_t	meshNameHeaderCount;
	uint32_t	unknown6C;						// +
	uint32_t	unknown70;						// | Always 0x00000000
	uint32_t	unknown74;						// | They might just be padding
	uint32_t	unknown78;						// |
	uint32_t	unknown7C;						// +
} wmb4_header_t;

/********** Vertex Data **********/

typedef struct wmb4_vertexHeader_s {
	uint32_t	vertexArrayOffset;	// see: wmb4_vertexXX_t
	uint32_t	unknown04;			// An offset to an array with vertexCount 8-byte entries of unknown use
	uint32_t	unknown08;			// 0
	uint32_t	unknown0C;			// 0
	uint32_t	vertexCount;		// Applies to both vertexArrayOffset and unknown04
	uint32_t	faceArrayOffset;	// see: wmb4_face_t
	uint32_t	faceCount;
} wmb4_vertexHeader_t;

// Note: The normals are packed in a 10.11.11 format as in the
// OpenGL Spec https://www.opengl.org/registry/doc/glspec45.core.pdf
typedef struct wmb4_vertex24_s {
	float		positionX;
	float		positionY;
	float		positionZ;
	uint16_t	textureU;	// Actually a half-precision float
	uint16_t	textureV;	// Actually a half-precision float
	uint32_t	normals;	// 10.11.11 packed normalized floats
	uint32_t	unknown14;
} wmb4_vertex24_t;

typedef struct wmb4_vertex28_s {
	float		positionX;
	float		positionY;
	float		positionZ;
	uint16_t	textureU;	// Actually a half-precision float
	uint16_t	textureV;	// Actually a half-precision float
	uint32_t	normals;	// 10.11.11 packed normalized floats
	uint32_t	unknown14;
	uint32_t	unknown18;
} wmb4_vertex28_t;

typedef struct wmb4_vertex32_s {
	float		positionX;
	float		positionY;
	float		positionZ;
	uint16_t	textureU;	// Actually a half-precision float
	uint16_t	textureV;	// Actually a half-precision float
	uint32_t	normals;	// 10.11.11 packed normalized floats
	uint32_t	unknown14;
	uint8_t		boneIndex0;		// +
	uint8_t		boneIndex1;		// | The index of the connected bone
	uint8_t		boneIndex2;		// |
	uint8_t		boneIndex3;		// +
	uint8_t		boneWeight0;	// +
	uint8_t		boneWeight1;	// | The weight of the connected bone
	uint8_t		boneWeight2;	// |
	uint8_t		boneWeight3;	// +
} wmb4_vertex32_t;

// The vertex indexes used for the faces are based on the first vertex
// in the mesh, NOT in the global vertex array.
// If you don't offset by the mesh's vertexStart you're
// going to have a bad time.
//
// Although all of the models appear to use tris, the faces don't group
// verticies in threes, so the format seems like it could support quads
// or n-gons too.
typedef uint16_t wmb4_face_t;

/********** Mesh Data **********/

typedef struct wmb4_mesh_s {
	uint32_t	vertexHeaderIndex;	// The index of the wmb4_vertexHeader_t to use
	uint32_t	vertexStart;		// The index of the first vertex in this mesh
	uint32_t	faceStart;			// The first face index in this mesh
	uint32_t	vertexCount;		// The number of verticies in the mesh
	uint32_t	faceCount;			// The number of face indexes in the mesh
} wmb4_mesh_t;


/*
 * In earlier revisions of the WMB format the second metadata array was for 
 * meshes that weren't directly rendered (there are indications that they were
 * used to generate shadows), but in MGR:R they appear to contain normal,
 * visble meshes.  There's probably more to it than that, but it's currently a
 * mystery.
 */
typedef struct wmb4_meshMetadataHeader_s {
	uint32_t	meshMetadataOffset;		// see: wmb4_meshMetadata_t
	uint32_t	meshMetadataCount;
	uint32_t	unknown08;				// 0x0000
	uint32_t	unknown0C;				// 0x0000
	uint32_t	unknown10;				// 0x0000
	uint32_t	unknown14;				// 0x0000
	uint32_t	altMeshMetadataOffset;	// see: wmb4_meshMetadata_t
	uint32_t	altMeshMetadataCount;
} wmb4_meshMetadataHeader_t;

typedef struct wmb4_meshMetadata_s {
	uint32_t	meshIndex;
	uint32_t	meshNameIndex;	// The index of the wmb4_meshName_t that applies to this mesh
	uint16_t	materialIndex;	// The index of the wmb4_material_t that applies to this mesh
	uint16_t	boneMapIndex;	// The index of the wmb4_boneMap_t that applies to this mesh
	uint32_t	unknown0C;		// Always seems to be 0x100
} wmb4_meshMetadata_t;

/********** Bone Data **********/

typedef struct wmb4_bone_s {
	uint16_t	unknown00;
	uint16_t	boneNumber;			// The bone's number (Display only! Don't use this for computing indexes or offsets or whatever)
	uint16_t	parentIndex;		// Index of the bone's parent in the global bone array (-1 if the bone has no parent)
	uint16_t	unknown06;			// Almost always 0x00
	float		positionRelativeX;	// + 
	float		positionRelativeY;	// | The bone's position relative to its parent (or the origin if it doesn't have a parent)
	float		positionRelativeZ;	// + 
	float		positionAbsoluteX;
	float		positionAbsoluteY;
	float		positionAbsoluteZ;
} wmb4_bone_t;

typedef struct wmb4_boneMap_s {
	uint32_t	boneMapOffset;		// see: wmb4_boneMapEntry_t
	uint32_t	boneMapCount;
} wmb4_boneMap_t;

// The index of the bone being mapped in the global bone array
typedef uint8_t wmb4_boneMapEntry_t;

/********** Materials Data **********/

typedef struct wmb4_material_s {
	uint32_t	shaderNameOffset;		// Offset to a null-terminated string that may be the name of the shader to use (observed values: "ois00_sbxxX", "cnm20_SbXXX")
	uint32_t	textureMapArrayOffset;	// see: wmb4_textureMap_t
	uint32_t	unknown08;				// 0x00000000
	uint32_t	matrixOffset;			// Offset to an array of floats; appears to be a matrix
	uint16_t	unknown10;
	uint16_t	unknown12;
	uint16_t	unknown14;
	uint16_t	unknown16;
} wmb4_material_t;

// Each entry appears to be an offset into the global texture identifier array for a different texture map
typedef struct wmb4_textureMap_s {
	uint32_t	unknown00;
	uint32_t	albedoMapTextureIndex;	// Albedo
	uint32_t	unknown08;
	uint32_t	unknown0C;
	uint32_t	unknown10;
	uint32_t	unknown14;
	uint32_t	unknown18;
	uint32_t	normalMapTextureIndex;	// Normal
} wmb4_textureMap_t;

typedef struct wmb4_textureIdentifier_s {
	uint32_t	unknown00;			// Some sort of flags?
	uint32_t	textureIdentifier;	// A global texture identifier (corresponds with the identifiers given in the wta file)
} wmb4_textureIdentifier_t;

/********** Mesh Names **********/

typedef struct wmb4_meshNameHeader_s {
	uint32_t	nameOffset;	// Offset to a null-terminated string
	uint32_t	unknown04;
	float		unknown08;
	float		unknown0C;
	float		unknown10;
	float		unknown14;
	float		unknown18;
	uint32_t	meshIdArrayOffset;	// 0x1C Offset to an array of shorts which indicate which meshes this names applies to (seems redundant with wmb4_meshMetadata_t)
	uint32_t	meshIdCount;
	uint32_t	unknown24;	// 0x00000000
	uint32_t	unknown28;	// 0x00000000
	uint32_t	unknown2C;	// 0x00000000
	uint32_t	unknown30;	// 0x00000000
	uint32_t	unknown3C;	// 0x00000000
	uint32_t	unknown40;	// 0x00000000
	uint32_t	unknown44;	// Offset to an array of shorts which look like some sort of mapping
	uint32_t	unknown48;	// Count
} wmb4_meshNameHeader_t;
